@@ -1501,14 +1501,12 @@ Yap_InitMemory(UInt Trail, UInt Heap, UInt Stack)
 #if RANDOMIZE_START_ADDRESS
   srand(time(NULL));
   UInt x = (rand()% 100)*YAP_ALLOC_SIZE ;
-  fprintf(stderr,"x=%lx\n", (unsigned long int)x);
   pm += x;
 #endif
   addr = InitWorkSpace(pm);
 #if RANDOMIZE_START_ADDRESS
   addr = (char *)addr+x;
   pm -= x;
-  fprintf(stderr,"addr=%p\n", addr);
 #endif
 
   InitHeap(addr);
