@@ -477,6 +477,8 @@ X_API Term    STD_PROTO(YAP_GetFromSlot,(Int));
 X_API Term   *STD_PROTO(YAP_AddressFromSlot,(Int));
 X_API void    STD_PROTO(YAP_PutInSlot,(Int, Term));
 X_API int     STD_PROTO(YAP_RecoverSlots,(int));
+X_API Int     STD_PROTO(YAP_ArgsToSlots,(int));
+X_API void    STD_PROTO(YAP_SlotsToArgs,(int, Int));
 X_API void    STD_PROTO(YAP_Throw,(Term));
 X_API void    STD_PROTO(YAP_Halt,(int));
 X_API Term   *STD_PROTO(YAP_TopOfLocalStack,(void));
@@ -493,9 +495,7 @@ X_API Term     STD_PROTO(YAP_CurrentModule,(void));
 X_API Term     STD_PROTO(YAP_CreateModule,(Atom));
 X_API Term     STD_PROTO(YAP_StripModule,(Term, Term *));
 X_API int      STD_PROTO(YAP_ThreadSelf,(void));
-X_API int      STD_PROTO(YAP_GetThreadRefCount,(int));
-X_API void     STD_PROTO(YAP_SetThreadRefCount,(int,int));
-X_API CELL     STD_PROTO(YAP_ThreadCreateEngine,(struct thread_attr_struct *));
+X_API int      STD_PROTO(YAP_ThreadCreateEngine,(struct thread_attr_struct *));
 X_API int      STD_PROTO(YAP_ThreadAttachEngine,(int));
 X_API int      STD_PROTO(YAP_ThreadDetachEngine,(int));
 X_API int      STD_PROTO(YAP_ThreadDestroyEngine,(int));
@@ -505,15 +505,22 @@ X_API Term     STD_PROTO(YAP_TermNil,(void));
 X_API int      STD_PROTO(YAP_AtomGetHold,(Atom));
 X_API int      STD_PROTO(YAP_AtomReleaseHold,(Atom));
 X_API Agc_hook STD_PROTO(YAP_AGCRegisterHook,(Agc_hook));
+X_API int      STD_PROTO(YAP_HaltRegisterHook,(HaltHookFunc, void *));
 X_API char    *STD_PROTO(YAP_cwd,(void));
 X_API Term     STD_PROTO(YAP_OpenList,(int));
 X_API Term     STD_PROTO(YAP_ExtendList,(Term, Term));
 X_API int      STD_PROTO(YAP_CloseList,(Term, Term));
 X_API int      STD_PROTO(YAP_IsAttVar,(Term));
 X_API Term     STD_PROTO(YAP_AttsOfVar,(Term));
-X_API Term     STD_PROTO(YAP_TermHash,(Term));
 X_API int      STD_PROTO(YAP_FileNoFromStream,(Term));
 X_API void    *STD_PROTO(YAP_FileDescriptorFromStream,(Term));
+X_API void    *STD_PROTO(YAP_Record,(Term));
+X_API Term     STD_PROTO(YAP_Recorded,(void *));
+X_API int      STD_PROTO(YAP_Erase,(void *));
+X_API int      STD_PROTO(YAP_Variant,(Term, Term));
+X_API int      STD_PROTO(YAP_ExactlyEqual,(Term, Term));
+X_API Int      STD_PROTO(YAP_TermHash,(Term, Int, Int, int));
+X_API int      STD_PROTO(YAP_SetYAPFlag,(yap_flag_t, int));
 
 static int (*do_getf)(void);
 
@@ -1052,6 +1059,46 @@ YAP_Unify(Term t1, Term t2)
   return out;
 }
 
+/* == */
+X_API int
+YAP_ExactlyEqual(Term t1, Term t2)
+{
+  int out;
+  BACKUP_MACHINE_REGS();
+
+  out = Yap_eq(t1, t2);
+
+  RECOVER_MACHINE_REGS();
+  return out;
+}
+
+/* =@= */
+X_API int
+YAP_Variant(Term t1, Term t2)
+{
+  int out;
+  BACKUP_MACHINE_REGS();
+
+  out = Yap_Variant(Deref(t1), Deref(t2));
+
+  RECOVER_MACHINE_REGS();
+  return out;
+}
+
+/* =@= */
+X_API Int
+YAP_TermHash(Term t, Int sz, Int depth, int variant)
+{
+  Int out;
+
+  BACKUP_MACHINE_REGS();
+
+  out = Yap_TermHash(t, sz, depth, variant);
+
+  RECOVER_MACHINE_REGS();
+  return out;
+}
+
 X_API Int
 YAP_CurrentSlot(void)
 {
@@ -1955,6 +2002,7 @@ YAP_RestartGoal(void)
     Yap_PrologMode = UserCCallMode;
     if (out == FALSE) {
       /* cleanup */
+      Yap_CloseSlots();
       Yap_trust_last();
       Yap_AllowRestart = FALSE;
     }
@@ -2039,7 +2087,28 @@ YAP_GoalHasException(Term *t)
   int out = FALSE;
   BACKUP_MACHINE_REGS();
   if (EX) {
-    *t = EX;
+    do {
+      Yap_Error_TYPE = YAP_NO_ERROR;
+      *t = Yap_FetchTermFromDB(EX);
+      if (Yap_Error_TYPE == YAP_NO_ERROR) {
+	RECOVER_MACHINE_REGS();
+	return TRUE;
+      } else if (Yap_Error_TYPE == OUT_OF_ATTVARS_ERROR) {
+	Yap_Error_TYPE = YAP_NO_ERROR;
+	if (!Yap_growglobal(NULL)) {
+	  Yap_Error(OUT_OF_ATTVARS_ERROR, TermNil, Yap_ErrorMessage);
+	  RECOVER_MACHINE_REGS();
+	  return FALSE;
+	}
+      } else {
+	Yap_Error_TYPE = YAP_NO_ERROR;
+	if (!Yap_growstack(EX->NOfCells*CellSize)) {
+	  Yap_Error(OUT_OF_STACK_ERROR, TermNil, Yap_ErrorMessage);
+	  RECOVER_MACHINE_REGS();
+	  return FALSE;
+	}
+      }
+    } while (*t == (CELL)0);
     out = TRUE;
   }
   RECOVER_MACHINE_REGS();
@@ -2050,7 +2119,11 @@ X_API void
 YAP_ClearExceptions(void)
 {
   Yap_ResetExceptionTerm();
-  EX = 0L;
+  if (EX) {
+    BallTerm = EX;
+  }
+  EX = NULL;
+  Yap_ResetExceptionTerm();    
   UncaughtThrow = FALSE;
 }
 
@@ -2087,7 +2160,7 @@ YAP_Read(int (*mygetc)(void))
   BACKUP_MACHINE_REGS();
 
   do_getf = mygetc;
-  sno = Yap_GetFreeStreamDForReading();
+  sno =  Yap_GetFreeStreamDForReading();
   if (sno < 0) {
     Yap_Error(SYSTEM_ERROR,TermNil, "new stream not available for YAP_Read");
     return TermNil;
@@ -2096,6 +2169,7 @@ YAP_Read(int (*mygetc)(void))
   Stream[sno].status |= Tty_Stream_f;
   tokstart = Yap_tokptr = Yap_toktide = Yap_tokenizer(sno, &tpos);
   Stream[sno].status = Free_Stream_f;
+  UNLOCK(Stream[sno].streamlock);
   if (Yap_ErrorMessage)
     {
       Yap_clean_tokenizer(tokstart, Yap_VarTable, Yap_AnonVarTable);
@@ -2312,7 +2386,7 @@ YAP_Init(YAP_init_args *yap_init)
   Yap_argc = yap_init->Argc;
 #if !BOOT_FROM_SAVED_STATE
   if (yap_init->SavedState) {
-    fprintf(stderr,"[ WARNING: threaded Yap will ignore saved state %s ]\n",yap_init->SavedState);
+    fprintf(stderr,"[ WARNING: threaded YAP will ignore saved state %s ]\n",yap_init->SavedState);
     yap_init->SavedState = NULL;
   }    
 #endif
@@ -2385,71 +2459,71 @@ YAP_Init(YAP_init_args *yap_init)
     yap_flags[QUIET_MODE_FLAG] = TRUE;
   }
 
-{ BACKUP_MACHINE_REGS();
-  Yap_InitYaamRegs();
+  { BACKUP_MACHINE_REGS();
+    Yap_InitYaamRegs();
 
 #if HAVE_MPI
-  Yap_InitMPI ();
+    Yap_InitMPI ();
 #endif
 #if HAVE_MPE
-  Yap_InitMPE ();
+    Yap_InitMPE ();
 #endif
-
-  if (yap_init->YapPrologRCFile != NULL) {
-    /*
-      This must be done before restore, otherwise
-      restore will print out messages ....
-    */
-    yap_flags[HALT_AFTER_CONSULT_FLAG] = yap_init->HaltAfterConsult;
-  }
-  /* tell the system who should cope with interruptions */
-  Yap_ExecutionMode = yap_init->ExecutionMode;
-  if (do_bootstrap) {
-    restore_result = YAP_BOOT_FROM_PROLOG;
-  } else if (BOOT_FROM_SAVED_STATE) {
-    restore_result = Yap_Restore(yap_init->SavedState, yap_init->YapLibDir);
-    if (restore_result == FAIL_RESTORE) {
-      yap_init->ErrorNo = Yap_Error_TYPE;
-      yap_init->ErrorCause = Yap_ErrorMessage;
-      /* shouldn't RECOVER_MACHINE_REGS();  be here ??? */
-      return YAP_BOOT_ERROR;
+    
+    if (yap_init->YapPrologRCFile != NULL) {
+      /*
+	This must be done before restore, otherwise
+	restore will print out messages ....
+      */
+      yap_flags[HALT_AFTER_CONSULT_FLAG] = yap_init->HaltAfterConsult;
     }
-  } else {
-    restore_result = YAP_BOOT_FROM_PROLOG;
-  }
-  yap_flags[FAST_BOOT_FLAG] = yap_init->FastBoot;
+    /* tell the system who should cope with interruptions */
+    Yap_ExecutionMode = yap_init->ExecutionMode;
+    if (do_bootstrap) {
+      restore_result = YAP_BOOT_FROM_PROLOG;
+    } else if (BOOT_FROM_SAVED_STATE) {
+      restore_result = Yap_Restore(yap_init->SavedState, yap_init->YapLibDir);
+      if (restore_result == FAIL_RESTORE) {
+	yap_init->ErrorNo = Yap_Error_TYPE;
+	yap_init->ErrorCause = Yap_ErrorMessage;
+	/* shouldn't RECOVER_MACHINE_REGS();  be here ??? */
+	return YAP_BOOT_ERROR;
+      }
+    } else {
+      restore_result = YAP_BOOT_FROM_PROLOG;
+    }
+    yap_flags[FAST_BOOT_FLAG] = yap_init->FastBoot;
 #if defined(YAPOR) || defined(TABLING)
 #ifdef TABLING
-  /* make sure we initialise this field */
-  GLOBAL_root_dep_fr = NULL;
+    /* make sure we initialise this field */
+    GLOBAL_root_dep_fr = NULL;
 #endif
-  make_root_frames();
+    make_root_frames();
 #ifdef YAPOR
-  init_workers();
+    init_workers();
 #endif /* YAPOR */
-  Yap_init_local();
+    Yap_init_local();
 #ifdef YAPOR
-  if (worker_id != 0) {
-#if SBA
-    /*
-      In the SBA we cannot just happily inherit registers
-      from the other workers
-    */
-    Yap_InitYaamRegs();
+    if (worker_id != 0) {
+#if SBA||ENV_COPY
+      /*
+	In the SBA we cannot just happily inherit registers
+	from the other workers
+      */
+      Yap_InitYaamRegs();
 #endif /* SBA */
 #ifndef THREADS
-    Yap_InitPreAllocCodeSpace();
+      Yap_InitPreAllocCodeSpace();
 #endif
-    /* slaves, waiting for work */
-    CurrentModule = USER_MODULE;
-    P = GETWORK_FIRST_TIME;
-    Yap_exec_absmi(FALSE);
-    Yap_Error(INTERNAL_ERROR, TermNil, "abstract machine unexpected exit (YAP_Init)");
-  }
+      /* slaves, waiting for work */
+      CurrentModule = USER_MODULE;
+      P = GETWORK_FIRST_TIME;
+      Yap_exec_absmi(FALSE);
+      Yap_Error(INTERNAL_ERROR, TermNil, "abstract machine unexpected exit (YAP_Init)");
+    }
 #endif /* YAPOR */
 #endif /* YAPOR || TABLING */
-  RECOVER_MACHINE_REGS();
- }
+    RECOVER_MACHINE_REGS();
+  }
   /* make sure we do this after restore */
   if (yap_init->MaxStackSize) {
     Yap_AllowLocalExpansion = FALSE;
@@ -2546,7 +2620,7 @@ YAP_Init(YAP_init_args *yap_init)
     }
     Yap_PutValue(Yap_FullLookupAtom("$live"), MkAtomTerm (Yap_FullLookupAtom("$true")));
   }
-  return restore_result;
+  return YAP_BOOT_FROM_PROLOG;
 }
 
 X_API Int
@@ -2811,27 +2885,27 @@ YAP_StripModule(Term t,  Term *modp)
 X_API int
 YAP_ThreadSelf(void)
 {
-#if USE_THREADS
+#if THREADS
   return Yap_thread_self();
 #else
-  return 0;
-#endif
-} 
-
-X_API CELL
-YAP_ThreadCreateEngine(struct thread_attr_struct *attr)
-{
-#if USE_THREADS
-  return Yap_thread_create_engine(attr);
-#else
-  return FALSE;
+  return -2;
 #endif
 } 
 
 X_API int
-YAP_ThreadAttachEngine(int wid)
+YAP_ThreadCreateEngine(struct thread_attr_struct * attr)
 {
-#if USE_THREADS
+#if THREADS
+  return Yap_thread_create_engine(attr);
+#else
+  return -1;
+#endif
+} 
+
+X_API int
+YAP_ThreadAttachEngine( int wid)
+{
+#if THREADS
   return Yap_thread_attach_engine(wid);
 #else
   return FALSE;
@@ -2841,7 +2915,7 @@ YAP_ThreadAttachEngine(int wid)
 X_API int
 YAP_ThreadDetachEngine(int wid)
 {
-#if USE_THREADS
+#if THREADS
   return Yap_thread_detach_engine(wid);
 #else
   return FALSE;
@@ -2851,7 +2925,7 @@ YAP_ThreadDetachEngine(int wid)
 X_API int
 YAP_ThreadDestroyEngine(int wid)
 {
-#if USE_THREADS
+#if THREADS
   return Yap_thread_destroy_engine(wid);
 #else
   return FALSE;
@@ -2884,6 +2958,12 @@ YAP_AGCRegisterHook(Agc_hook hook)
   return old;
 } 
 
+X_API int
+YAP_HaltRegisterHook(HaltHookFunc hook, void * closure)
+{
+  return Yap_HaltRegisterHook(hook, closure);
+} 
+
 X_API char *
 YAP_cwd(void)
 {
@@ -2967,20 +3047,6 @@ YAP_AttsOfVar(Term t)
   return attv->Atts;
 }
 
-X_API Term
-YAP_TermHash(Term t)
-{
-  attvar_record *attv;
-  
-  t = Deref(t);
-  if (!IsVarTerm(t))
-    return TermNil;
-  if (IsAttVar(VarOfTerm(t)))
-    return TermNil;
-  attv = (attvar_record *)VarOfTerm(t);
-  return attv->Atts;
-}
-
 X_API int
 YAP_FileNoFromStream(Term t)
 {
@@ -3001,3 +3067,69 @@ YAP_FileDescriptorFromStream(Term t)
   return Yap_FileDescriptorFromStream(t);
 }
 
+X_API void *
+YAP_Record(Term t)
+{
+ 
+  return (void *)Yap_StoreTermInDB(Deref(t), 0);
+}
+
+X_API Term
+YAP_Recorded(void *handle)
+{
+  return Yap_FetchTermFromDB((DBTerm *)handle);
+}
+
+X_API int
+YAP_Erase(void *handle)
+{
+  Yap_ReleaseTermFromDB((DBTerm *)handle);
+  return 1;
+}
+
+X_API Int
+YAP_ArgsToSlots(int n)
+{
+  Int slot = Yap_NewSlots(n);
+  CELL *ptr0 = LCL0+slot, *ptr1=&ARG1;
+  while (n--) {
+    *ptr0++ = *ptr1++;
+  }
+  return slot;
+}
+
+X_API void
+YAP_SlotsToArgs(int n, Int slot)
+{
+  CELL *ptr0 = LCL0+slot, *ptr1=&ARG1;
+  while (n--) {
+    *ptr1++ = *ptr0++;
+  }
+}
+
+
+X_API int
+YAP_SetYAPFlag(yap_flag_t flag, int val)
+{
+  switch (flag) {
+  case YAPC_ENABLE_GC:
+    if (val) {
+      Yap_PutValue(AtomGc, MkAtomTerm(AtomTrue));
+    } else {
+      Yap_PutValue(AtomGc, TermNil);
+    }
+    return TRUE;
+  case YAPC_ENABLE_AGC:
+    if (val) {
+      AGcThreshold = 10000;
+    } else {
+      AGcThreshold = 0;
+    }
+    return TRUE;
+  default:
+    return FALSE;
+  }
+}
+
+
+
