@@ -28,6 +28,9 @@
 #if HAVE_STRING_H
 #include <string.h>
 #endif
+#if THREADS && !defined(TABLING)
+#include <opt.mavar.h>
+#endif
 
 #if !HAVE_STRNCAT
 #define strncat(s0,s1,sz)   strcat(s0,s1)
@@ -210,7 +213,7 @@ MoveLocalAndTrail(void)
 #endif
 }
 
-#if defined(THREADS) && defined(YAPOR)
+#ifdef THREADS
 
 static void
 CopyLocalAndTrail(void)
@@ -235,8 +238,6 @@ IncrementalCopyStacksFromWorker(void)
 	 (size_t) (LOCAL_end_trail_copy - LOCAL_start_trail_copy));
 }
 
-#include "opt.mavar.h"
-
 static CELL
 worker_p_binding(int worker_p, CELL *aux_ptr)
 {
@@ -281,8 +282,9 @@ RestoreTrail(int worker_p)
     } else if (IsPairTerm(aux_cell)) {
       /* avoid frozen segments */
       aux_cell = (CELL) RepPair(aux_cell);
-      if ((ADDR) aux_cell >= TrailBase)
+      if (IN_BETWEEN(Yap_TrailBase, aux_cell, Yap_TrailTop)) {
         aux_tr = (tr_fr_ptr) aux_cell;
+      }
 #endif /* TABLING */
 #ifdef MULTI_ASSIGNMENT_VARIABLES
     } else if (IsApplTerm(aux_cell)) {
@@ -1071,6 +1073,7 @@ fix_compiler_instructions(PInstr *pcpc)
       break;
       /* hopefully nothing to do */
     case nop_op:
+    case ensure_space_op:
     case get_atom_op:
     case put_atom_op:
     case get_num_op:
@@ -1429,7 +1432,12 @@ Yap_growglobal(CELL **ptr)
 {
   unsigned long sz = sizeof(CELL) * K16;
 
-#if defined(YAPOR) || defined(THREADS)
+#if defined(YAPOR) && !defined(THREADS)
+  if (number_workers != 1) {
+    Yap_Error(OUT_OF_STACK_ERROR,TermNil,"cannot grow Global: more than a worker/thread running");
+    return(FALSE);
+  }
+#elif defined(THREADS)
   if (NOfThreads != 1) {
     Yap_Error(OUT_OF_STACK_ERROR,TermNil,"cannot grow Global: more than a worker/thread running");
     return(FALSE);
@@ -1867,9 +1875,9 @@ Yap_CopyThreadStacks(int worker_q, int worker_p, int incremental)
     LOCAL_end_local_copy = 
       (CELL)PtoLocAdjust((CELL *)LOCAL_end_local_copy);
     LOCAL_start_trail_copy = 
-      (CELL)PtoTRAdjust((CELL *)LOCAL_start_trail_copy);
+      (CELL)PtoTRAdjust((tr_fr_ptr)LOCAL_start_trail_copy);
     LOCAL_end_trail_copy = 
-      (CELL)PtoTRAdjust((CELL *)LOCAL_end_trail_copy);
+      (CELL)PtoTRAdjust((tr_fr_ptr)LOCAL_end_trail_copy);
     AdjustStacksAndTrail(0, STACK_INCREMENTAL_COPYING);
     RestoreTrail(worker_p);
     TR = (tr_fr_ptr) LOCAL_end_trail_copy;
