@@ -23,6 +23,7 @@ static char     SccsId[] = "%W% %G%";
  *
  */
 
+#include <stdlib.h>
 #include "Yap.h"
 #include "yapio.h"
 #include "alloc.h"
@@ -88,6 +89,8 @@ struct restore_info rinfo[MAX_THREADS];
 
 struct thread_globs Yap_thread_gl[MAX_THREADS];
 
+pthread_t Yap_master_thread;
+
 #else
 
 struct restore_info rinfo;
@@ -163,8 +166,6 @@ int      Yap_CritLocks = 0;
 
 /********* streams ********************************************/
 
-int Yap_c_input_stream, Yap_c_output_stream, Yap_c_error_stream;
-
 YP_FILE *Yap_stdin;
 YP_FILE *Yap_stdout;
 YP_FILE *Yap_stderr;
@@ -406,8 +407,8 @@ static Opdef    Ops[] = {
   {"\\/", yfx, 500},
   {"><", yfx, 500},
   {"#", yfx, 500},
-  {"xor", yfx, 400},
   {"rdiv", yfx, 400},
+  {"div", yfx, 400},
   {"*", yfx, 400},
   {"/", yfx, 400},
   {"//", yfx, 400},
@@ -1135,7 +1136,8 @@ InitInvisibleAtoms(void)
 }
 
 #ifdef  THREADS
-InitThreadHandle(wid)
+static void
+InitThreadHandle(int wid)
 {
     FOREIGN_ThreadHandle(wid).in_use = FALSE;
     FOREIGN_ThreadHandle(wid).zombie = FALSE;
@@ -1144,9 +1146,26 @@ InitThreadHandle(wid)
     FOREIGN_ThreadHandle(wid).thread_inst_count = 0LL;
 #endif
     pthread_mutex_init(&(FOREIGN_ThreadHandle(wid).tlock), NULL);  
+    pthread_mutex_init(&(FOREIGN_ThreadHandle(wid).tlock_status), NULL);  
+    FOREIGN_ThreadHandle(wid).tdetach = (CELL)0;
+    FOREIGN_ThreadHandle(wid).cmod = (CELL)0;
 }
 #endif
 
+static void
+InitSWIBuffers(int wid)
+{
+  int i;
+
+  FOREIGN_WL(wid)->SWI_buffers_[0] = malloc(SWI_BUF_SIZE);
+  FOREIGN_WL(wid)->SWI_buffers_sz_[0] = SWI_BUF_SIZE;
+  for (i=1; i <= SWI_BUF_RINGS; i++) {
+    FOREIGN_WL(wid)->SWI_buffers_[i] = NULL;
+    FOREIGN_WL(wid)->SWI_buffers_sz_[i] = 0;
+  }
+}
+
+
 static void 
 InitFirstWorkerThreadHandle(void)
 {
@@ -1156,7 +1175,6 @@ InitFirstWorkerThreadHandle(void)
   ThreadHandle.default_yaam_regs = 
     &Yap_standard_regs;
   ThreadHandle.pthread_handle = pthread_self();
-  ThreadHandle.pthread_handle = pthread_self();
   pthread_mutex_init(&FOREIGN_ThreadHandle(0).tlock, NULL);
   pthread_mutex_init(&FOREIGN_ThreadHandle(0).tlock_status, NULL);
   ThreadHandle.tdetach = MkAtomTerm(AtomFalse);
@@ -1182,15 +1200,21 @@ Yap_CloseScratchPad(void)
 #include "iglobals.h"
 
 #if defined(YAPOR) || defined(THREADS)
-#define MAX_INITS 1
+#define MAX_INITS MAX_AGENTS
 #else
 #define MAX_INITS 1
 #endif
 
+#if defined(YAPOR) &&  !defined(THREADS)
+struct worker_shared *Yap_global;
+#else
 struct worker_shared Yap_Global;
+#endif
 
-#if defined(YAPOR) || defined(THREADS)
-struct worker_local	Yap_WLocal[MAX_WORKERS];
+#if defined(YAPOR) &&  !defined(THREADS)
+struct worker_local	*Yap_WLocal;
+#elif defined(YAPOR) || defined(THREADS)
+struct worker_local	Yap_WLocal[MAX_AGENTS];
 #else
 struct worker_local	Yap_WLocal;
 #endif
@@ -1201,8 +1225,9 @@ InitCodes(void)
   int wid;
 #include "ihstruct.h"
   InitGlobal();
-  for (wid = 0; wid < MAX_INITS; wid++)
+  for (wid = 0; wid < MAX_INITS; wid++) {
     InitWorker(wid);
+  }
   InitFirstWorkerThreadHandle();
   /* make sure no one else can use these two atoms */
   CurrentModule = 0;
@@ -1240,11 +1265,11 @@ Yap_InitWorkspace(UInt Heap, UInt Stack, UInt Trail, UInt Atts, UInt max_table_s
   int             i;
 
   /* initialise system stuff */
-
 #if PUSH_REGS
 #ifdef THREADS
   pthread_key_create(&Yap_yaamregs_key, NULL);
   pthread_setspecific(Yap_yaamregs_key, (const void *)&Yap_standard_regs);
+  Yap_master_thread = pthread_self();
 #else
   /* In this case we need to initialise the abstract registers */
   Yap_regp = &Yap_standard_regs;
@@ -1317,6 +1342,33 @@ Yap_InitWorkspace(UInt Heap, UInt Stack, UInt Trail, UInt Atts, UInt max_table_s
   }
 }
 
+int
+Yap_HaltRegisterHook (HaltHookFunc f, void * env)
+{
+  struct halt_hook *h;
+
+  if (!(h = (struct halt_hook *)Yap_AllocCodeSpace(sizeof(struct halt_hook))))
+    return FALSE;
+  h->environment = env;
+  h->hook = f;
+  LOCK(BGL);
+  h->next = Yap_HaltHooks;
+  Yap_HaltHooks = h;
+  UNLOCK(BGL);
+  return TRUE;
+}
+
+static void
+run_halt_hooks(int code)
+{
+  struct halt_hook *hooke = Yap_HaltHooks;
+
+  while (hooke) {
+    hooke->hook(code, hooke->environment);
+    hooke = hooke->next;
+  }
+}
+
 void
 Yap_exit (int value)
 {
@@ -1324,15 +1376,17 @@ Yap_exit (int value)
   unmap_memory();
 #endif /* YAPOR */
 
+  if (! (Yap_PrologMode & BootMode) ) {
 #ifdef LOW_PROF
-  remove("PROFPREDS");
-  remove("PROFILING");
+    remove("PROFPREDS");
+    remove("PROFILING");
 #endif
 #if defined MYDDAS_MYSQL || defined MYDDAS_ODBC
-  Yap_MYDDAS_delete_all_myddas_structs();
+    Yap_MYDDAS_delete_all_myddas_structs();
 #endif
-  if (! (Yap_PrologMode & BootMode) )
+    run_halt_hooks(value);
     Yap_ShutdownLoadForeign();
+  }
   exit(value);
 }
 
