@@ -110,6 +110,7 @@ STATIC_PROTO(int   get_coded, (int, OPCODE []));
 STATIC_PROTO(void  restore_codes, (void));
 STATIC_PROTO(Term  AdjustDBTerm, (Term, Term *));
 STATIC_PROTO(void  RestoreDB, (DBEntry *));
+STATIC_PROTO(void  RestoreDBTerm, (DBTerm *, int));
 STATIC_PROTO(void  CleanClauses, (yamop *, yamop *,PredEntry *));
 STATIC_PROTO(void  rehash, (CELL *, int, int));
 STATIC_PROTO(void  CleanCode, (PredEntry *));
@@ -404,7 +405,7 @@ save_regs(int mode)
       return -1;
     if (putout(CreepFlag) < 0)
       return -1;
-    if (putout(EX) < 0)
+    if (putcellptr((CELL *)EX) < 0)
       return -1;
 #if defined(SBA) || defined(TABLING)
     if (putcellptr(H_FZ) < 0)
@@ -607,7 +608,12 @@ do_save(int mode) {
 static Int 
 p_save(void)
 {
-#if defined(YAPOR) || defined(THREADS)
+#if defined(YAPOR) && !defined(THREADS)
+  if (number_workers != 1) {
+    Yap_Error(SYSTEM_ERROR,TermNil,"cannot perform save: more than a worker/thread running");
+    return(FALSE);
+  }
+#elif defined(THREADS)
   if (NOfThreads != 1) {
     Yap_Error(SYSTEM_ERROR,TermNil,"cannot perform save: more than a worker/thread running");
     return(FALSE);
@@ -621,7 +627,13 @@ p_save(void)
 static Int 
 p_save2(void)
 {
-#if defined(YAPOR) || defined(THREADS)
+#if defined(YAPOR) && !defined(THREADS)
+  if (number_workers != 1) {
+    Yap_Error(SYSTEM_ERROR,TermNil,
+	       "cannot perform save: more than a worker/thread running");
+    return(FALSE);
+  }
+#elif defined(THREADS)
   if (NOfThreads != 1) {
     Yap_Error(SYSTEM_ERROR,TermNil,
 	       "cannot perform save: more than a worker/thread running");
@@ -756,6 +768,7 @@ get_heap_info(void)
   if (Yap_ErrorMessage)
       return -1;
   OldHeapTop = (ADDR) get_cellptr();
+
   if (Yap_ErrorMessage)
       return -1;
   OldHeapUsed = (Int) get_cell();
@@ -841,7 +854,7 @@ get_regs(int flag)
     CreepFlag = get_cell();
     if (Yap_ErrorMessage)
       return -1;
-    EX = get_cell();
+    EX = (struct DB_TERM *)get_cellptr();
     if (Yap_ErrorMessage)
       return -1;
 #if defined(SBA) || defined(TABLING)
@@ -939,8 +952,9 @@ get_hash(void)
 static int 
 CopyCode(void)
 {
-  if (myread(splfild, (char *) Yap_HeapBase, (Unsigned(OldHeapTop) - Unsigned(OldHeapBase))) < 0)
+  if (myread(splfild, (char *) Yap_HeapBase, (Unsigned(OldHeapTop) - Unsigned(OldHeapBase))) < 0) {
     return -1;
+  }
   return 1;
 }
 
@@ -986,7 +1000,7 @@ static int
 get_coded(int flag, OPCODE old_ops[])
 {
   char my_end_msg[256];
-
+  
   if (get_regs(flag) < 0)
     return -1;
   if (get_insts(old_ops) < 0)
@@ -1046,8 +1060,10 @@ restore_regs(int flag)
     HB = PtoLocAdjust(HB);
     YENV = PtoLocAdjust(YENV);
     S = PtoGloAdjust(S);
-    if (EX)
-      EX = AbsAppl(PtoGloAdjust(RepAppl(EX)));
+    if (EX) {
+      EX = DBTermAdjust(EX);
+      RestoreDBTerm(EX, TRUE);
+    }
     WokenGoals = AbsAppl(PtoGloAdjust(RepAppl(WokenGoals)));
   }
 }
@@ -1383,7 +1399,7 @@ commit_to_saved_state(char *s, CELL *Astate, CELL *ATrail, CELL *AStack, CELL *A
    */
   errout = Yap_stderr;
 #endif
-  return(mode);
+  return mode;
 }
 
 static void
@@ -1398,10 +1414,27 @@ cat_file_name(char *s, char *prefix, char *name, unsigned int max_length)
   strncat(s, name, max_length-1);
 }
 
+static int try_open(char *inpf, CELL *Astate, CELL *ATrail, CELL *AStack, CELL *AHeap, char *buf) {
+  int mode;
+
+  
+  if ((splfild = open_file(inpf, O_RDONLY)) < 0) {
+    return FAIL_RESTORE;
+  }
+  if (buf[0] == '\0')
+    strncpy(buf, inpf, YAP_FILENAME_MAX);
+  if ((mode = commit_to_saved_state(inpf,Astate,ATrail,AStack,AHeap)) != FAIL_RESTORE) {
+    Yap_ErrorMessage = NULL;
+    return mode;
+  }
+  return mode;
+}
+
 static int 
 OpenRestore(char *inpf, char *YapLibDir, CELL *Astate, CELL *ATrail, CELL *AStack, CELL *AHeap)
 {
   int mode = FAIL_RESTORE;
+  char save_buffer[YAP_FILENAME_MAX+1];
 
   //  Yap_ErrorMessage = NULL;
   if (inpf == NULL) {
@@ -1410,21 +1443,25 @@ OpenRestore(char *inpf, char *YapLibDir, CELL *Astate, CELL *ATrail, CELL *AStac
 #endif
       inpf = StartUpFile;
   }
+  /* careful it starts from the root */
+  if (inpf[0] != '/') {
 #if __simplescalar__
-  /* does not implement getcwd */
-  strncpy(Yap_FileNameBuf,yap_pwd,YAP_FILENAME_MAX);
+    /* does not implement getcwd */
+    strncpy(Yap_FileNameBuf,yap_pwd,YAP_FILENAME_MAX);
 #elif HAVE_GETCWD
-  if (getcwd (Yap_FileNameBuf, YAP_FILENAME_MAX) == NULL)
-    Yap_FileNameBuf[0] = '\0';
+    if (getcwd (Yap_FileNameBuf, YAP_FILENAME_MAX) == NULL)
+      Yap_FileNameBuf[0] = '\0';
 #else
-  if (getwd (Yap_FileNameBuf) == NULL)
-    Yap_FileNameBuf[0] = '\0';
+    if (getwd (Yap_FileNameBuf) == NULL)
+      Yap_FileNameBuf[0] = '\0';
 #endif
-  strncat(Yap_FileNameBuf, "/", YAP_FILENAME_MAX-1);
-  strncat(Yap_FileNameBuf, inpf, YAP_FILENAME_MAX-1);
+    strncat(Yap_FileNameBuf, "/", YAP_FILENAME_MAX-1);
+    strncat(Yap_FileNameBuf, inpf, YAP_FILENAME_MAX-1);
+  } else {
+    strncat(Yap_FileNameBuf, inpf, YAP_FILENAME_MAX-1);
+  }
   if (inpf != NULL && (splfild = open_file(inpf, O_RDONLY)) > 0) {
-    if ((mode = commit_to_saved_state(inpf,Astate,ATrail,AStack,AHeap)) != FAIL_RESTORE) {
-      Yap_ErrorMessage = NULL;
+    if ((mode = try_open(inpf,Astate,ATrail,AStack,AHeap,save_buffer)) != FAIL_RESTORE) {
       return mode;
     }
   }
@@ -1435,11 +1472,12 @@ OpenRestore(char *inpf, char *YapLibDir, CELL *Astate, CELL *ATrail, CELL *AStac
     */
     if (YapLibDir != NULL) {
       cat_file_name(Yap_FileNameBuf, Yap_LibDir, inpf, YAP_FILENAME_MAX);
-      if ((splfild = open_file(Yap_FileNameBuf, O_RDONLY)) > 0) {
-	if ((mode = commit_to_saved_state(Yap_FileNameBuf,Astate,ATrail,AStack,AHeap)) != FAIL_RESTORE) {
-	  Yap_ErrorMessage = NULL;
-	  return mode;
-	}
+      if ((mode = try_open(Yap_FileNameBuf,Astate,ATrail,AStack,AHeap,save_buffer)) != FAIL_RESTORE) {
+	return mode;
+      }
+    } else {
+      if ((mode = try_open(Yap_FileNameBuf,Astate,ATrail,AStack,AHeap,save_buffer)) != FAIL_RESTORE) {
+	return mode;
       }
     }
 #if HAVE_GETENV
@@ -1447,11 +1485,8 @@ OpenRestore(char *inpf, char *YapLibDir, CELL *Astate, CELL *ATrail, CELL *AStac
       char *yap_env = getenv("YAPLIBDIR");
       if (yap_env != NULL) {
 	cat_file_name(Yap_FileNameBuf, yap_env, inpf, YAP_FILENAME_MAX);
-	if ((splfild = open_file(Yap_FileNameBuf, O_RDONLY)) > 0) {
-	  if ((mode = commit_to_saved_state(Yap_FileNameBuf,Astate,ATrail,AStack,AHeap)) != FAIL_RESTORE) {
-	    Yap_ErrorMessage = NULL;
-	    return mode;
-	  }
+	if ((mode = try_open(Yap_FileNameBuf,Astate,ATrail,AStack,AHeap,save_buffer)) != FAIL_RESTORE) {
+	  return mode;
 	}
       }
     }
@@ -1459,9 +1494,8 @@ OpenRestore(char *inpf, char *YapLibDir, CELL *Astate, CELL *ATrail, CELL *AStac
     if (YAP_LIBDIR != NULL) {
       cat_file_name(Yap_FileNameBuf, YAP_LIBDIR, inpf, YAP_FILENAME_MAX);
       if ((splfild = open_file(Yap_FileNameBuf, O_RDONLY)) > 0) {
-	if ((mode = commit_to_saved_state(Yap_FileNameBuf,Astate,ATrail,AStack,AHeap)) != FAIL_RESTORE) {
-	  Yap_ErrorMessage = NULL;
-	  return(mode);
+	if ((mode = try_open(Yap_FileNameBuf,Astate,ATrail,AStack,AHeap,save_buffer)) != FAIL_RESTORE) {
+	  return mode;
 	}
       }
     }
@@ -1500,11 +1534,8 @@ OpenRestore(char *inpf, char *YapLibDir, CELL *Astate, CELL *ATrail, CELL *AStac
       pt[1] = '\0';
       strncat(Yap_FileNameBuf,"lib/Yap/startup.yss",YAP_FILENAME_MAX);
     }
-    if ((splfild = open_file(Yap_FileNameBuf, O_RDONLY)) > 0) {
-      if ((mode = commit_to_saved_state(Yap_FileNameBuf,Astate,ATrail,AStack,AHeap)) != FAIL_RESTORE) {
-	Yap_ErrorMessage = NULL;
-	return(mode);
-      }
+    if ((mode = try_open(Yap_FileNameBuf,Astate,ATrail,AStack,AHeap,save_buffer)) != FAIL_RESTORE) {
+      return mode;
     }
   }
  end:
@@ -1512,19 +1543,13 @@ OpenRestore(char *inpf, char *YapLibDir, CELL *Astate, CELL *ATrail, CELL *AStac
   /* try to open from current directory */
   /* could not open file */
   if (Yap_ErrorMessage == NULL) {
-#if __simplescalar__
-    /* does not implement getcwd */
-    strncpy(Yap_FileNameBuf,yap_pwd,YAP_FILENAME_MAX);
-#elif HAVE_GETCWD
-    if (getcwd (Yap_FileNameBuf, YAP_FILENAME_MAX) == NULL)
-      Yap_FileNameBuf[0] = '\0';
-#else
-    if (getwd (Yap_FileNameBuf) == NULL)
-      Yap_FileNameBuf[0] = '\0';
-#endif
-    strncat(Yap_FileNameBuf, "/", YAP_FILENAME_MAX-1);
-    strncat(Yap_FileNameBuf, inpf, YAP_FILENAME_MAX-1);
-    do_system_error(PERMISSION_ERROR_OPEN_SOURCE_SINK,"could not open saved state");
+    if (save_buffer[0]) {
+      strncpy(Yap_FileNameBuf, save_buffer, YAP_FILENAME_MAX-1);
+      do_system_error(PERMISSION_ERROR_OPEN_SOURCE_SINK,"incorrect saved state");
+    } else {
+      strncpy(Yap_FileNameBuf, inpf, YAP_FILENAME_MAX-1);
+      do_system_error(PERMISSION_ERROR_OPEN_SOURCE_SINK,"could not open saved state");
+    }
   }
   return FAIL_RESTORE;
 }
@@ -1750,7 +1775,12 @@ p_restore(void)
   int mode;
 
   Term t1 = Deref(ARG1);
-#if defined(YAPOR) || defined(THREADS)
+#if defined(YAPOR) && !defined(THREADS)
+  if (number_workers != 1) {
+    Yap_Error(SYSTEM_ERROR,TermNil,"cannot perform save: more than a worker/thread running");
+    return(FALSE);
+  }
+#elif defined(THREADS)
   if (NOfThreads != 1) {
     Yap_Error(SYSTEM_ERROR,TermNil,"cannot perform save: more than a worker/thread running");
     return(FALSE);
