@@ -123,6 +123,90 @@ wrputws(wchar_t *s, wrf writewch)		/* writes a string	 */
     wrputc(*s++, writewch);
 }
 
+#ifdef USE_GMP
+
+static char *
+ensure_space(size_t sz) {
+  char *s;
+
+  s = (char *) Yap_PreAllocCodeSpace();
+  while (s+sz >= (char *)AuxSp) {
+#if USE_SYSTEM_MALLOC
+    /* may require stack expansion */
+    if (!Yap_ExpandPreAllocCodeSpace(sz, NULL, TRUE)) {
+      s = NULL;
+      break;
+    }
+    s = (char *) Yap_PreAllocCodeSpace();
+#else
+    s = NULL;
+#endif
+  }
+  if (!s) {
+    s = (char *)TR;
+    while (s+sz >= Yap_TrailTop) {
+      if (!Yap_growtrail(sz/sizeof(CELL), FALSE)) {
+	s = NULL;
+	break;
+      }
+      s = (char *)TR;
+    }
+  }
+  if (!s) {
+    s = (char *)H;
+    if (s+sz >= (char *)ASP) {
+      Yap_Error(OUT_OF_STACK_ERROR,TermNil,"not enough space to write bignum: it requires %d bytes", sz);
+      s = NULL;
+    }
+  }
+  return s;
+}
+
+static void
+write_mpint(MP_INT *big, wrf writewch) {
+  char *s;
+
+  s = ensure_space(3+mpz_sizeinbase(big, 10));
+  if (mpz_sgn(big) < 0) {
+    if (lastw == symbol)
+      wrputc(' ', writewch);  
+  } else {
+    if (lastw == alphanum)
+      wrputc(' ', writewch);
+  }
+  if (!s) {
+    s = mpz_get_str(NULL, 10, big);
+    if (!s)
+      return;
+    wrputs(s,writewch);
+    free(s);
+  } else {
+    mpz_get_str(s, 10, big);
+    wrputs(s,writewch);
+  }
+}
+#endif
+
+	/* writes a bignum	 */
+static void 
+writebig(Term t, int p, int depth, int rinfixarg, struct write_globs *wglb, struct rewind_term *rwt)
+{
+#ifdef USE_GMP
+  CELL *pt = RepAppl(t)+1;
+  if (pt[0] == BIG_INT) 
+  {
+    MP_INT *big = Yap_BigIntOfTerm(t);
+    write_mpint(big, wglb->writewch);
+    return;
+  } else if (pt[0] == BIG_RATIONAL) {
+    Term trat = Yap_RatTermToApplTerm(t);
+    writeTerm(trat, p, depth, rinfixarg, wglb, rwt);
+    return;
+  }
+#endif
+  wrputs("0",wglb->writewch);
+}	                  
+
 static void 
 wrputf(Float f, wrf writewch)		/* writes a float	 */
 	                  
@@ -572,7 +656,7 @@ writeTerm(Term t, int p, int depth, int rinfixarg, struct write_globs *wglb, str
     putAtom(Atom3Dots, wglb->Quote_illegal, wglb->writewch);
     return;
   }
-  if (EX != 0)
+  if (EX)
     return;
   t = Deref(t);
   if (IsVarTerm(t)) {
@@ -584,17 +668,17 @@ writeTerm(Term t, int p, int depth, int rinfixarg, struct write_globs *wglb, str
   } else if (IsPairTerm(t)) {
     if (wglb->Use_portray) {
       Term targs[1];
-      Term old_EX = 0L;
+      struct DB_TERM *old_EX = NULL;
       Int sl = 0;
 
       targs[0] = t;
       Yap_PutValue(AtomPortray, MkAtomTerm(AtomNil));
-      if (EX != 0L) old_EX = EX;
+      if (EX) old_EX = EX;
       sl = Yap_InitSlot(t);      
       Yap_execute_goal(Yap_MkApplTerm(FunctorPortray, 1, targs), 0, 1);
       t = Yap_GetFromSlot(sl);
       Yap_RecoverSlots(1);
-      if (old_EX != 0L) EX = old_EX;
+      if (old_EX != NULL) EX = old_EX;
       if (Yap_GetValue(AtomPortray) == MkAtomTerm(AtomTrue))
 	return;
     }
@@ -629,64 +713,9 @@ writeTerm(Term t, int p, int depth, int rinfixarg, struct write_globs *wglb, str
       case (CELL)FunctorLongInt:
 	wrputn(LongIntOfTerm(t),wglb->writewch);
 	return;
-      case (CELL)FunctorBigInt:
-#ifdef USE_GMP
-	{
-	  MP_INT *big = Yap_BigIntOfTerm(t);
-	  char *s;
-	  s = (char *) Yap_PreAllocCodeSpace();
-	  while (s+3+mpz_sizeinbase(big, 10) >= (char *)AuxSp) {
-#if USE_SYSTEM_MALLOC
-	    /* may require stack expansion */
-	    if (!Yap_ExpandPreAllocCodeSpace(3+mpz_sizeinbase(big, 10), NULL, TRUE)) {
-	      s = NULL;
-	      break;
-	    }
-	    s = (char *) Yap_PreAllocCodeSpace();
-#else
-	    s = NULL;
-#endif
-	  }
-	  if (!s) {
-	    s = (char *)TR;
-	    while (s+3+mpz_sizeinbase(big, 10) >= Yap_TrailTop) {
-	      if (!Yap_growtrail((3+mpz_sizeinbase(big, 10))/sizeof(CELL), FALSE)) {
-		s = NULL;
-		break;
-	      }
-	      s = (char *)TR;
-	    }
-	  }
-	  if (!s) {
-	    s = (char *)H;
-	    if (s+3+mpz_sizeinbase(big, 10) >= (char *)ASP) {
-	      Yap_Error(OUT_OF_STACK_ERROR,TermNil,"not enough space to write bignum: it requires %d bytes", 3+mpz_sizeinbase(big, 10));
-	      s = NULL;
-	    }
-	  }
-	  if (mpz_sgn(big) < 0) {
-	    if (lastw == symbol)
-	      wrputc(' ', wglb->writewch);  
-	  } else {
-	    if (lastw == alphanum)
-	      wrputc(' ', wglb->writewch);
-	  }
-	  if (!s) {
-	    s = mpz_get_str(NULL, 10, big);
-	    if (!s)
-	      return;
-	    wrputs(s,wglb->writewch);
-	    free(s);
-	  } else {
-	    mpz_get_str(s, 10, big);
-	    wrputs(s,wglb->writewch);
-	  }
-	}
-#else
-	{
-	  wrputs("0",wglb->writewch);
-	}
-#endif
+	/* case (CELL)FunctorBigInt: */
+      default:
+	writebig(t, p, depth, rinfixarg, wglb, rwt);
 	return;
       }
     }
@@ -731,18 +760,18 @@ writeTerm(Term t, int p, int depth, int rinfixarg, struct write_globs *wglb, str
 #endif
     if (wglb->Use_portray) {
       Term targs[1];
-      Term old_EX = 0L;
+      struct DB_TERM *old_EX = NULL;
       Int sl = 0;
 
       targs[0] = t;
       Yap_PutValue(AtomPortray, MkAtomTerm(AtomNil));
-      if (EX != 0L) old_EX = EX;
+      if (EX) old_EX = EX;
       sl = Yap_InitSlot(t);      
       Yap_execute_goal(Yap_MkApplTerm(FunctorPortray, 1, targs),0, 1);
       t = Yap_GetFromSlot(sl);
       Yap_RecoverSlots(1);
-      if (old_EX != 0L) EX = old_EX;
-      if (Yap_GetValue(AtomPortray) == MkAtomTerm(AtomTrue) || EX != 0L)
+      if (old_EX) EX = old_EX;
+      if (Yap_GetValue(AtomPortray) == MkAtomTerm(AtomTrue) || EX)
 	return;
     }
     if (!wglb->Ignore_ops &&
