@@ -14,7 +14,7 @@
 *************************************************************************/
 
 #include "config.h"
-#if defined(ENV_COPY) || defined(TABLING)
+#if defined(ENV_COPY) || defined(TABLING) || defined(THREADS)
 #include "opt.config.h"
 #endif /* YAPOR || TABLING */
 
@@ -40,7 +40,7 @@
 #error Do not define multiple or-parallel models
 #endif /* (ENV_COPY && (ACOW || SBA)) || (ACOW && SBA) */
 
-#if defined(ENV_COPY) || defined(ACOW) || defined(SBA)
+#if defined(ENV_COPY) || defined(ACOW) || defined(SBA) || defined(THREADS)
 #define YAPOR 1
 #endif /* ENV_COPY || ACOW || SBA */
 
@@ -90,7 +90,7 @@
 #undef  USE_THREADED_CODE
 #endif
 #define inline __inline
-#define YAP_VERSION "Yap-6.0.5"
+#define YAP_VERSION "YAP-6.2.0"
 
 #define BIN_DIR "c:\\Yap\\bin"
 #define LIB_DIR "c:\\Yap\\lib\\Yap"
@@ -276,7 +276,7 @@ extern char Yap_Option[20];
 #endif
 #endif /* !IN_SECOND_QUADRANT */
 
-/* #define RANDOMIZE_START_ADDRESS 1*/
+/* #define RANDOMIZE_START_ADDRESS 1 */
 
 #ifdef USE_SYSTEM_MALLOC
 #define HEAP_INIT_BASE  0L
@@ -351,7 +351,7 @@ typedef CELL Term;
 
 #if !defined(YAPOR) && !defined(THREADS)
 #include <nolocks.h>
-#elif USE_PTHREAD_LOCKING || defined(__APPLE__) || defined(__CYGWIN__)
+#elif USE_PTHREAD_LOCKING || defined(__CYGWIN__)
 
 #ifndef _XOPEN_SOURCE
 #define _XOPEN_SOURCE 600
@@ -361,7 +361,7 @@ typedef CELL Term;
 typedef pthread_mutex_t lockvar;
 typedef pthread_rwlock_t rwlock_t;
 
-#elif defined(i386) || defined(__x86_64__)
+#elif defined(i386)|| defined(__x86_64__)
 typedef volatile int lockvar;
 #include <locks_x86.h>
 #elif defined(sparc) || defined(__sparc)
@@ -423,10 +423,10 @@ typedef pthread_rwlock_t rwlock_t;
 #ifdef __alpha
 #include <locks_alpha_funcs.h>
 #endif
-#ifdef YAPOR
-#define MAX_AGENTS MAX_WORKERS
-#elif defined(THREADS)
+#if defined(THREADS)
 #define MAX_AGENTS MAX_THREADS
+#elif defined(YAPOR)
+#define MAX_AGENTS MAX_WORKERS
 #endif
 #endif
 
@@ -683,7 +683,7 @@ typedef enum
 	if you place things in the lower addresses (power to the libc people).
 */
 
-#if (defined(_AIX) || (defined(__APPLE__) && !defined(__LP64__)) || defined(_WIN32) || defined(sparc) || defined(__sparc) || defined(mips) || defined(__FreeBSD__) || defined(_POWER) || defined(__POWERPC__) || defined(__linux__) || defined(IN_SECOND_QUADRANT) || defined(__CYGWIN__))
+#if (defined(_AIX) || (defined(__APPLE__) && !defined(__LP64__)) || defined(_WIN32) || defined(sparc) || defined(__sparc) || defined(mips) || defined(__FreeBSD__) || defined(_POWER) || defined(__POWERPC__) || defined(__linux__) || defined(IN_SECOND_QUADRANT) || defined(__CYGWIN__)) || defined(__NetBSD__)
 #define USE_LOW32_TAGS 1
 #endif
 
@@ -773,7 +773,6 @@ typedef struct thread_globs
 
 extern struct thread_globs Yap_thread_gl[MAX_THREADS];
 
-
 #define    Yap_LocalBase  Yap_thread_gl[worker_id].local_base
 #define    Yap_GlobalBase Yap_thread_gl[worker_id].global_base
 #define    Yap_TrailBase  Yap_thread_gl[worker_id].trail_base
@@ -784,6 +783,10 @@ extern struct thread_globs Yap_thread_gl[MAX_THREADS];
 #define    Yap_Error_Size   Yap_thread_gl[worker_id].error_size
 #define    Yap_ErrorSay    Yap_thread_gl[worker_id].error_say
 #define    Yap_RestartEnv    Yap_thread_gl[worker_id].restart_env
+
+/* This is the guy who actually started the system, and who has the correct registers */
+extern pthread_t Yap_master_thread;
+
 #else
 extern ADDR Yap_HeapBase,
   Yap_LocalBase, Yap_GlobalBase, Yap_TrailBase, Yap_TrailTop;
@@ -1305,4 +1308,27 @@ typedef enum
     COMPILE_ALL          /* compile all predicates */
   } yap_exec_mode;
 
+/********* slots ***********************/
+
+
+static inline void
+Yap_StartSlots(void) {
+  *--ASP = MkIntegerTerm(CurSlot);
+  *--ASP = MkIntTerm(0);
+  CurSlot = LCL0-ASP;
+}
+
+static inline void
+Yap_CloseSlots(void) {
+  Int old_slots;
+  old_slots = IntOfTerm(ASP[0]);
+  ASP += (old_slots+1);
+  CurSlot = IntegerOfTerm(*ASP);
+  ASP++;
+}
+
+static inline Int
+Yap_CurrentSlot(void) {
+  return IntOfTerm(ASP[0]);
+}
 
