@@ -1520,4 +1520,14 @@ PredPropByAtom (Atom at, Term cur_mod)
   return Yap_NewPredPropByAtom (ae, cur_mod);
 }
 
+#if DEBUG_PELOCKING
+#define PELOCK(I,Z)						\
+  { LOCK((Z)->PELock); (Z)->StatisticsForPred.NOfEntries=(I);(Z)->StatisticsForPred.NOfHeadSuccesses=pthread_self(); }
+#define UNLOCKPE(I,Z)						\
+  ( (Z)->StatisticsForPred.NOfRetries=(I),  UNLOCK((Z)->PELock) )
+#else
+#define PELOCK(I,Z) LOCK((Z)->PELock)
+#define UNLOCKPE(I,Z)	UNLOCK((Z)->PELock)
+#endif
+
 #endif
