@@ -868,7 +868,7 @@ in use by Prolog. This can't work as   the  table is guarded by a mutex.
 Now we use the FD_CLOEXEC flag in Snew();
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-void
+X_API void
 PL_cleanup_fork(void)
 { stopItimer();
 }
@@ -4307,9 +4307,24 @@ BeginPredDefs(file)
 EndPredDefs
 
 #if __YAP_PROLOG__
+static pl_Sgetc(IOSTREAM *s)
+{
+  return Sgetc(s);
+}
+
 static void
 init_yap_extras()
 {
+  swi_io_struct swiio;
+
+  swiio.f = FUNCTOR_dstream1;
+  swiio.get_c = pl_Sgetc;
+  swiio.put_c = Sputc;
+  swiio.get_w = Sgetcode;
+  swiio.put_w = Sputcode;
+  swiio.flush_s = Sflush;
+  swiio.close_s = Sclose;
+  PL_YAP_InitSWIIO(&swiio);
   initCharTypes();
   initFiles();
   initGlob();
